
/* clang-print.cc */

#include "clang-print.h"

#include "str.h"
#include "conv.h"

#include "clang/AST/ASTConsumer.h"
#include "clang/AST/DeclFriend.h"

// #include "clang/Basic/LLVM.h"

OPEN_NAMESPACE

using llvm::dyn_cast;

/* ---------------------------------------------------------------------- */

TScanner::TScanner ()
{
   fCtx = NULL;
   fLastDecl = 0;
}

TScanner::~TScanner ()
{
}

void TScanner::PutQualifier (const string txt)
{
   Put (txt); // !?
   Put (" ");
}

void TScanner::ShowInfo (string msg, string location)
{
   Info (msg + ", " + location);
}

void TScanner::ShowWarning (string msg, string location)
{
   Warning (msg + ", " + location);
}

/********************************* NUMBERS ********************************/

inline size_t APIntToSize (const llvm::APInt& num)
{
   return *num.getRawData ();
}

inline long APIntToLong (const llvm::APInt& num)
{
   return *num.getRawData ();
}

inline string APIntToStr (const llvm::APInt& num)
{
   return num.toString (10, true);
}

/******************************** UNKNOWN *********************************/

string TScanner::GetSrcLocation (clang::SourceLocation L)
{
   #if 0
   string location = "";
   llvm::raw_string_ostream stream (location);
   clang::SourceManager& source_manager = fCtx->getSourceManager ();
   L.print (stream, source_manager);
   return stream.str ();
   #else
      return "";
   #endif
}

string TScanner::GetLocation (clang::Decl * D)
{
   if (D == NULL)
   {
      return "";
   }
   else
   {
      #if 0
      string location = "";
      llvm::raw_string_ostream stream (location);
      clang::SourceManager& source_manager = fCtx->getSourceManager ();
      D->getLocation ().print (stream, source_manager);
      return stream.str ();
      #else
         return "";
      #endif
   }
}

string TScanner::GetName (clang::Decl * D)
{
   assert (D != NULL);

   string name = "";
   // string kind = D->getDeclKindName ();

   if (clang::NamedDecl * ND = dyn_cast <clang::NamedDecl> (D))
   {
      name = ND->getQualifiedNameAsString ();
   }

   return name;
}

void TScanner::DeclInfo (clang::Decl * D)
{
   assert (D != NULL);

   string location = GetLocation (D);
   string kind = D->getDeclKindName ();
   string name = GetName (D);
   ShowInfo ("Scan: " + kind + " declaration " + name, location);
}

void TScanner::UnknownDecl (clang::Decl * D, string txt)
{
   assert (D != NULL);

   // unknown - this kind of declaration was not known to programmer
   string location = GetLocation (D);
   string kind = D->getDeclKindName ();
   string name = GetName (D);
   if (txt != "")
      txt = txt + " ";
   ShowWarning ("Unknown " + txt + kind + " declaration " + name, location);
}

void TScanner::UnknownType (clang::QualType qual_type)
{
   string location = GetLocation (fLastDecl);
   string kind = qual_type.getTypePtr ()->getTypeClassName ();
   ShowWarning ("Unknown " + kind + " type " + qual_type.getAsString (), location);
}

/********************************* TYPES **********************************/

void TScanner::ScanPointerType (const clang::PointerType * T)
{
   assert (T != NULL);

   ScanQualType (T->getPointeeType ());
   Put (" * ");
}

void TScanner::ScanLValueReferenceType (const clang::LValueReferenceType * T)
{
   assert (T != NULL);

   ScanQualType (T->getPointeeType ());
   Put (" & ");
}

void TScanner::ScanRValueReferenceType (const clang::RValueReferenceType * T)
{
   assert (T != NULL);

   ScanQualType (T->getPointeeType ());
   Put (" && ");
}

void TScanner::ScanMemberPointerType (const clang::MemberPointerType * T)
{
   assert (T != NULL);

   // !? Put (T->getClass()->getQualifiedNameAsString ());
   Put ("::*");
   ScanQualType (T->getPointeeType ());
}

void TScanner::ScanArrayType (const clang::ArrayType * T)
{
   assert (T != NULL);

   ScanQualType (T->getElementType ());
   Put (" [");

   if (const clang::ConstantArrayType * CT = dyn_cast <clang::ConstantArrayType> (T))
   {
      size_t size = APIntToSize (CT->getSize ()) ;
      Put (IntToStr (size)); // !?
   }

   Put ("]");
}

void TScanner::ScanFunctionType (const clang::FunctionType * T)
{
   assert (T != NULL);

   // result type

   ScanQualType (T->getResultType ());

   // parameters

   Put ("(");

   bool any = false;

   if (const clang::FunctionProtoType * FT = dyn_cast <clang::FunctionProtoType> (T))
   {
      for (clang::FunctionProtoType::arg_type_iterator I = FT->arg_type_begin (), E = FT->arg_type_end (); I != E; ++I)
      {
         if (any) Put (", ");
         any = true;
         ScanQualType (*I);
      }
   }

   Put (")");
}

#if 0
   void TScanner::ScanEnumType (clang::EnumType * T) { }
   void TScanner::ScanRecordType (clang::RecordType * T) { }
   void TScanner::ScanElaboratedType (clang::ElaboratedType * T) { }
   void TScanner::ScanTypedefType (clang::TypedefType * T) { }
   void TScanner::ScanBuiltinType (clang::BuiltinType * T) { }
   void TScanner::ScanInjectedClassNameType (clang::InjectedClassNameType * T) { }
   void TScanner::ScanDependentNameType (clang::DependentNameType * T) { }
   void TScanner::ScanTemplateTypeParmType (clang::TemplateTypeParmType * T) { }
   void TScanner::ScanSubstTemplateTypeParmType (clang::SubstTemplateTypeParmType * T) { }
   void TScanner::ScanTemplateSpecializationType (clang::TemplateSpecializationType * T) { }
#endif

void TScanner::ScanQualType (clang::QualType qual_type)
{
   // const and volatile

   if (qual_type.isConstQualified ())
      PutQualifier ("const");

   if (qual_type.isVolatileQualified ())
      PutQualifier ("volatile");

   const clang::Type * T = qual_type.getTypePtr ();

   // see clang/include/clang/AST/TypeNodes.def

   switch (T->getTypeClass ())
   {
      case clang::Type::Pointer:
         ScanPointerType (dyn_cast <clang::PointerType> (T));
         break;

      case clang::Type::MemberPointer:
         ScanMemberPointerType (dyn_cast <clang::MemberPointerType> (T));
         break;

      case clang::Type::LValueReference:
         ScanLValueReferenceType (dyn_cast <clang::LValueReferenceType> (T));
         break;

      case clang::Type::RValueReference:
         ScanRValueReferenceType (dyn_cast <clang::RValueReferenceType> (T));
         break;

      case clang::Type::ConstantArray:
      case clang::Type::IncompleteArray:
      case clang::Type::DependentSizedArray:
      case clang::Type::VariableArray:
         ScanArrayType (dyn_cast <clang::ArrayType> (T));
         break;

      case clang::Type::FunctionProto:
      case clang::Type::FunctionNoProto:
         ScanFunctionType (dyn_cast <clang::FunctionType> (T));
         break;

      case clang::Type::Elaborated:
      case clang::Type::Builtin:
      case clang::Type::Complex:
      case clang::Type::Enum:
      case clang::Type::Record:
         StyledText (qual_type.getAsString (), type_style);
         break;

      case clang::Type::InjectedClassName:
      case clang::Type::DependentName:
      case clang::Type::SubstTemplateTypeParm:
      case clang::Type::TemplateTypeParm:
      case clang::Type::TemplateSpecialization:
         StyledText (qual_type.getAsString (), unknown_style /* special_style */);
         break;

      default:
         StyledText (qual_type.getAsString (), unknown_style);
         UnknownType (qual_type);
         break;
   }
}

/******************************* EXPRESSION *******************************/

void TScanner::ScanExpr (clang::Expr * expr)
{
   assert (expr != NULL);

   #if 0
   clang::LangOptions lang_opts;
   clang::PrintingPolicy print_opts (lang_opts); // !?

   std::string text = "";
   llvm::raw_string_ostream stream (text);

   expr->printPretty (stream, NULL, print_opts);

   Put (stream.str ());
   #endif
}

/******************************** VARIABLE ********************************/

void TScanner::ScanStorageClass (clang::VarDecl * D)
{
   assert (D != NULL);

   // if (D->isStaticDataMember ())
   //    PutQualifier ("static");

   switch (D->getStorageClass ())
   {
      case clang::SC_Auto:
         PutQualifier ("auto");
         break;

      case clang::SC_Register:
         PutQualifier ("register");
         break;

      case clang::SC_Extern:
         PutQualifier ("extern");
         break;

      case clang::SC_Static:
         PutQualifier ("static");
         break;

      case clang::SC_PrivateExtern:
         PutQualifier ("extern"); // !?
         break;
   }
}

void TScanner::ScanVariable (clang::VarDecl * D)
{
   assert (D != NULL);

   StyledText ("/* variable */", variable_style);
   Put (" ");

   // modifiers

   ScanStorageClass (D);

   if (clang::FieldDecl * S = dyn_cast <clang::FieldDecl> (D))
   {
      if (S->isMutable ())
         PutQualifier ("mutable");
   }

   // type

   ScanQualType (D->getType ());

   // name

   std::string name = D->getNameAsString ();

   // initializer

   clang::Expr * init = D->getInit ();
   if (init != NULL)
   {
      Put (" = ");
      ScanExpr (init);
   }

   Put (" ");
   StyledText (name, variable_style);
   Put (";");
   PutEol ();
}

/******************************** FUNCTION ********************************/

void TScanner::ScanFuncModifiers (clang::FunctionDecl * D)
{
   switch (D->getStorageClass ())
   {
      case clang::SC_Extern:
         PutQualifier ("extern");
         break;

      case clang::SC_Static:
         PutQualifier ("static");
         break;
   }

   if (D->isInlineSpecified ())
       PutQualifier ("inline");

   if (clang::CXXMethodDecl * M = dyn_cast <clang::CXXMethodDecl> (D))
   {
      if (M->isVirtual ())
         PutQualifier ("virtual");

      if (M->isStatic ())
         PutQualifier("static");

      // I do not know, how to recognize void func () const
      // PutQualifier ("const");

      if (clang::CXXConstructorDecl * S = dyn_cast <clang::CXXConstructorDecl> (M))
      {
         PutQualifier ("constructor");

         if (S->isCopyConstructor ())
            PutQualifier ("copyconstructor");

         if (S->isExplicit ())
            PutQualifier ("explicit");

         if (S->isConvertingConstructor (false))  // !? parameter allow explicit
            PutQualifier("converter");

      }
      else if (clang::CXXDestructorDecl * S = dyn_cast <clang::CXXDestructorDecl> (M))
      {
         PutQualifier ("destructor");

      }
      else if (clang::CXXConversionDecl * S = dyn_cast <clang::CXXConversionDecl> (M))
      {
         PutQualifier ("converter");
         if (S->isExplicit ())
            PutQualifier ("explicit");
      }
   }
}

void TScanner::ScanFuncParameters (clang::FunctionDecl * D)
{
   assert (D != NULL);

   Put ("(");

   bool any = false;

   for (clang::FunctionDecl::param_iterator I = D->param_begin (), E = D->param_end (); I != E; ++I)
   {
      clang::ParmVarDecl * P = *I;

      if (any) Put (", ");
      any = true;

      // Put (P->getType ().getAsString ());
      ScanQualType (P->getType());
      Put (" ");
      Put (P->getNameAsString ());

      // NO if (P->hasDefaultArg ()) // check hasUnparsedDefaultArg () and hasUninstantiatedDefaultArg ()
      if (P->getInit ())
      {
         Put (" = ");
         ScanExpr (P->getDefaultArg ());
      }
   }

   if (D->isVariadic ())
   {
      if (any) Put (", ");
      Put ("... ");
   }

   Put (")");
}

void TScanner::ScanFuncExceptions (const clang::FunctionProtoType * D)
{
   if (D->hasExceptionSpec ())
   {
      Put (" ");
      Put ("throw");
      Put (" ");
      Put ("(");
      /*
      if (D->hasAnyExceptionSpec ())
         Put ("...");
      else
      */
      {
         int cnt = D->getNumExceptions ();
         for (int i = 0 ; i < cnt; i++)
         {
            ScanQualType (D->getExceptionType (i));
         }
      }
      Put (")");
   }
}

void TScanner::ScanConstructorInitializer (clang::FunctionDecl * D)
{
   if (clang::CXXConstructorDecl * S = dyn_cast <clang::CXXConstructorDecl> (D))
   {
      #if 0
      bool any = false;
      for (clang::CXXConstructorDecl::init_const_iterator I = S->init_begin(), E = S->init_end() ; I != E; ++I)
      {
         if (!any) Put (" : "); else Put (", ");
         any = true;

         clang::CXXCtorInitializer * init = *I;
         Put (init->getMember()->getNameAsString());
         Put (" ");
         Put ("(");
         ScanExpr (init->getInit ());
         Put (")");
      }
      #endif
   }
}

void TScanner::ScanFunction (clang::FunctionDecl * D)
{
   assert (D != NULL);

   // modifiers

   ScanFuncModifiers (D);

   // type

   ScanQualType (D->getResultType ());

   Put (" ");

   // name

   Put (D->getNameAsString ());

   // paramers

   Put (" ");

   ScanFuncParameters (D);

   // exceptions

   const clang::Type * T = D->getType().getTypePtr();
   const clang::FunctionProtoType * F = dyn_cast <clang::FunctionProtoType> (T);

   if (F != NULL)
      ScanFuncExceptions (F);

   // initializers

   ScanConstructorInitializer (D);

   // modifiers

   if (D->isPure ())
      Put (" = 0");

   if (F != NULL)
   {
      unsigned int type_quals = F->getTypeQuals ();

      if (type_quals & clang::Qualifiers::Const != 0)
          PutQualifier ("const");

      if (type_quals & clang::Qualifiers::Volatile != 0)
          PutQualifier ("volatile");

      if (type_quals & clang::Qualifiers::Restrict != 0)
          PutQualifier ("restrict");
   }

   Put (";");
   PutEol ();
}

/************************************ ENUM ************************************/

void TScanner::ScanEnum (clang::EnumDecl * D)
{
   assert (D != NULL);

   // name

   string name = D->getNameAsString ();

   StyledText ("enum", enum_style);
   Put (" ");
   StyledText (name, enum_style);
   PutEol ();

   StyledText ("{", enum_style);
   PutEol ();
   IncIndent ();

   // items

   bool any = false;

   for (clang::EnumDecl::enumerator_iterator I = D->enumerator_begin (), E = D->enumerator_end (); I != E; ++I)
   {
      std::string item_name = I->getNameAsString ();

      if (any)
      {
         StyledText (",", enum_style);
         PutEol ();
      }
      any = true;

      // name

      StyledText (item_name, enum_style);

      // value

      clang::Expr * init = I->getInitExpr ();
      if (init != NULL)
      {
         Put (" = ");
         ScanExpr (init);
      }
   }

   if (any)
      PutEol ();

   DecIndent ();

   StyledText ("};", enum_style);
   Put (" ");
   StyledText ("// end of " + name, comment_style);
   PutEol ();
}

/*********************************** CLASS ************************************/

void TScanner::ScanVisibilityModifiers (clang::AccessSpecifier access)
{
   DecIndent ();

   switch (access)
   {
      case clang::AS_private:
         PutLn ("private:");
         break;

      case clang::AS_protected:
         PutLn ("protected:");
         break;

      case clang::AS_public:
         PutLn ("public:");
         break;
   }

   IncIndent ();
}

void TScanner::ScanAccessSpecDecl (clang::AccessSpecDecl * D)
{
   ScanVisibilityModifiers (D->getAccess ());
}

void TScanner::ScanFriendDecl (clang::FriendDecl *D)
{
   assert (D != NULL);

   if (D->getFriendDecl() != NULL)
   {
      PutQualifier ("friend");
      Put (D->getFriendDecl()->getNameAsString());
      Put (";");
      PutEol ();
   }
}

void TScanner::ScanClass (clang::RecordDecl * D)
{
   assert (D != NULL);

   clang::TagDecl::TagKind kind = D->getTagKind ();

   if (kind == clang::TTK_Class)
      StyledText ("class", class_style);
   else if (kind == clang::TTK_Struct)
      StyledText ("struct", class_style);
   else if (kind == clang::TTK_Union)
      StyledText ("union", class_style);

   Put (" ");

   StyledText (D->getNameAsString (), class_style);

   // base classes

   bool any = false;

   if (clang::CXXRecordDecl * C = dyn_cast <clang::CXXRecordDecl> (D))
   {
      if (C->getDefinition () != NULL)  // important
         for (clang::CXXRecordDecl::base_class_iterator I = C->bases_begin (), E = C->bases_end (); I != E; ++I)
         {
            if (! any)
               Put (" : ");
            else
               Put (", ");

            any = true;

            // clang::CXXBaseSpecifier  b = *I;

            if (I->isVirtual ())
               StyledText ("virtual ", class_style);

            ScanVisibilityModifiers (I->getAccessSpecifier ());

            ScanQualType (I->getType ());
         }
   }

   StyledText ("{", class_style);
   PutEol ();

   IncIndent ();

   // members

   for (clang::DeclContext::decl_iterator I = D->decls_begin (), E = D->decls_end (); I != E; ++I)
      ScanDecl (*I);

   DecIndent ();

   StyledText ("};", class_style);
   Put (" ");
   StyledText ("// end of " + D->getNameAsString (), comment_style);
   PutEol ();
}

/******************************** TEMPLATE *********************************/

void TScanner::ScanTemplateParameter (clang::NamedDecl * D)
{
   assert (D != NULL);

   switch (D->getKind ())
   {
      case clang::Decl::TemplateTypeParm:
         {
            clang::TemplateTypeParmDecl * P = dyn_cast <clang::TemplateTypeParmDecl> (D);

            if (P->wasDeclaredWithTypename ())
               PutQualifier ("typename");
            else
               PutQualifier ("class");

            if (P->isParameterPack ())
               PutQualifier ("...");

            Put (P->getNameAsString ());

            if (P->hasDefaultArgument ())
            {
               Put (" = ");
               Put (P->getDefaultArgument ().getAsString ());
            }
         }
         break;

      case clang::Decl::NonTypeTemplateParm:
         {
            clang::NonTypeTemplateParmDecl * P = dyn_cast <clang::NonTypeTemplateParmDecl> (D);
            Put (P->getType ().getAsString ());

            if (clang::IdentifierInfo * N = P->getIdentifier ())
            {
               Put (" ");
               Put (N->getName ());
            }

            if (P->hasDefaultArgument ())
            {
               Put (" = ");
               ScanExpr (P->getDefaultArgument ());
            }
         }
         break;

      default:
         UnknownDecl (D, "template parameter");
   }
}

void TScanner::ScanTemplateParameterList (clang::TemplateParameterList * list)
{
   StyledText ("<", template_style);
   Put (" ");

   bool any = false;

   for (clang::TemplateParameterList::iterator I = list->begin (), E = list->end (); I != E; ++I)
   {
      if (any) Put (", ");
      any = true;

      ScanTemplateParameter (*I);
   }

   Put (" ");
   StyledText (">", template_style);
}

void TScanner::ScanTemplate (clang::TemplateDecl * D)
{
   assert (D != NULL);

   StyledText ("template", template_style);
   Put (" ");
   ScanTemplateParameterList (D->getTemplateParameters ());
   PutEol ();

   IncIndent ();
   ScanDecl (D->getTemplatedDecl ());
   DecIndent ();
}

/*********************************** USING ************************************/

void TScanner::ScanUsing (clang::UsingDecl * D)
{
   assert (D != NULL);

   StyledText ("using", namespace_style);
   Put (" ");

   #if 0
   std::string text = "";
   llvm::raw_string_ostream stream (text);
   clang::LangOptions LO;
   clang::PrintingPolicy PP (LO);
   D->getTargetNestedNameDecl()->print (stream, PP);
   Put (stream.str());
   #endif

   Put (";");
   PutEol ();
}

void TScanner::ScanUsingDirective (clang::UsingDirectiveDecl * D)
{
   assert (D != NULL);

   StyledText ("using namespace", namespace_style);
   Put (" ");
   Put (D->getNominatedNamespaceAsWritten()->getQualifiedNameAsString());
   Put (";");
   PutEol ();
}

void TScanner::ScanUnresolvedUsingValueMember (clang::UnresolvedUsingValueDecl * D)
{
   assert (D != NULL);

   StyledText ("using", namespace_style);
   Put (" ");

   #if 0
   std::string text = "";
   llvm::raw_string_ostream stream (text);
   clang::LangOptions LO;
   clang::PrintingPolicy PP (LO);
   D->getTargetNestedNameSpecifier()->print (stream, PP);
   Put (stream.str());
   #endif

   Put (";");
   PutEol ();
}

/*********************************** GLOBAL ************************************/

void TScanner::ScanNamespace (clang::NamespaceDecl * D)
{
   assert (D != NULL);

   std::string ns_name = D->getNameAsString (); // or  getQualifiedNameAsString

   StyledText ("namespace", namespace_style);
   Put (" ");
   StyledText (ns_name, namespace_style);
   PutEol ();

   StyledText ("{", namespace_style);
   PutEol ();

   IncIndent ();

   clang::DeclContext * DC = dyn_cast <clang::DeclContext> (D);
   if (DC != NULL)
      for (clang::DeclContext::decl_iterator I = DC->decls_begin (), E = DC->decls_end (); I != E; ++I)
         ScanDecl (*I);

   DecIndent ();

   StyledText ("}", namespace_style);
   Put (" ");
   StyledText ("// end of " + ns_name, comment_style);
   PutEol ();
}

void TScanner::ScanNamespaceAlias (clang::NamespaceAliasDecl * D)
{
   assert (D != NULL);

   StyledText ("namespace", namespace_style);
   Put (" ");
   Put (D->getNameAsString ());
   Put (" = ");
   Put (D->getAliasedNamespace ()->getQualifiedNameAsString ());
   Put (";");
   PutEol ();
}

void TScanner::ScanLinkageSpec (clang::LinkageSpecDecl * D)
{
   assert (D != NULL);

       PutQualifier ("extern");

   if (D->getLanguage() == clang::LinkageSpecDecl::lang_c)
      Put (QuoteStr ("C"));

   for (clang::DeclContext::decl_iterator I = D->decls_begin (), E = D->decls_end (); I != E; ++I)
      ScanDecl (*I);
}

void TScanner::ScanTypedef (clang::TypedefDecl * D)
{
   assert (D != NULL);

    StyledText ("typedef", type_style);
   Put (" ");

   // type

   ScanQualType (D->getUnderlyingType ());

   // name

   std::string name = D->getQualifiedNameAsString ();

   Put (" ");
   StyledText (name, type_style);

   Put (";");
   PutEol ();
}

void TScanner::ScanTranslationUnit (clang::TranslationUnitDecl * D)
{
   assert (D != NULL);

    clang::DeclContext * DC = clang::TranslationUnitDecl::castToDeclContext (D);
   for (clang::DeclContext::decl_iterator I = DC->decls_begin (), E = DC->decls_end (); I != E; ++I)
      ScanDecl (*I);
}

/************************************ DECL ************************************/

void TScanner::ScanDecl (clang::Decl * D)
{
   // see clang/include/clang/AST/DeclNodes.def
   // see clang/lib/AST/DeclPrinter.cpp

   fLastDecl = D; // only for debugging

   switch (D->getKind ())
   {

      case clang::Decl::Var:
      // !? case clang::Decl::Field:
         ScanVariable (dyn_cast <clang::VarDecl> (D));
         break;

      case clang::Decl::Function:
      case clang::Decl::CXXMethod:
      case clang::Decl::CXXConstructor:
      case clang::Decl::CXXDestructor:
      case clang::Decl::CXXConversion:
         ScanFunction (dyn_cast <clang::FunctionDecl> (D));
         break;

      case clang::Decl::Typedef:
         ScanTypedef (dyn_cast <clang::TypedefDecl> (D));
         break;

      case clang::Decl::Enum:
         ScanEnum (dyn_cast <clang::EnumDecl> (D));
         break;

      case clang::Decl::Record:
      case clang::Decl::CXXRecord:
         ScanClass (dyn_cast <clang::RecordDecl> (D));
         break;

      case clang::Decl::ClassTemplate:
      // case clang::Decl::ClassTemplateSpecialization:
      // case clang::Decl::ClassTemplatePartialSpecialization:
      case clang::Decl::FunctionTemplate:
         ScanTemplate (dyn_cast <clang::TemplateDecl> (D));
         break;

      case clang::Decl::TranslationUnit:
         ScanTranslationUnit (dyn_cast <clang::TranslationUnitDecl> (D));
         break;

      #if 0
      case clang::Decl::Namespace:
         ScanNamespace (dyn_cast <clang::NamespaceDecl> (D));
         break;
      #endif

      case clang::Decl::NamespaceAlias:
         ScanNamespaceAlias (dyn_cast <clang::NamespaceAliasDecl> (D));
         break;

      case clang::Decl::LinkageSpec:
         ScanLinkageSpec (dyn_cast <clang::LinkageSpecDecl> (D));
         break;

      case clang::Decl::Using:
         ScanUsing (dyn_cast <clang::UsingDecl> (D));
         break;

      case clang::Decl::UsingDirective:
         ScanUsingDirective (dyn_cast <clang::UsingDirectiveDecl> (D));
         break;

      // case clang::Decl::UsingShadow:
         // ScanUsingShadow (dyn_cast <clang::UsingShadowDecl> (D));
         // break;

      case clang::Decl::UnresolvedUsingValue:
         ScanUnresolvedUsingValueMember (dyn_cast <clang::UnresolvedUsingValueDecl> (D));
         break;

      case clang::Decl::Friend:
         ScanFriendDecl (dyn_cast <clang::FriendDecl> (D));
         break;

      case clang::Decl::AccessSpec:
         ScanAccessSpecDecl (dyn_cast <clang::AccessSpecDecl> (D));
         break;

      default:
         UnknownDecl (D);
   }
}

void TScanner::Scan (clang::ASTContext * C, clang::Decl * D)
{
   fCtx = C;
   ScanDecl (D);
}

/* -------------------------------------------------------------------------- */

CLOSE_NAMESPACE
